from __future__ import annotations

from json import dumps
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from github.Label import Label
from github.PullRequest import PullRequest
from github.Team import Team as GithubTeam

from bar_raiser.autofixes.notify_reviewer_teams import (
    LABEL_TO_REMOVE,
    ReviewRequest,
    create_slack_message,
    main,
    process_pull_request,
    process_review_request,
)


@pytest.fixture
def mock_pull_request() -> PullRequest:
    mock = MagicMock(spec=PullRequest)
    mock.user.login = "testuser"
    mock.html_url = "https://github.com/test/pr/1"
    mock.number = 1
    mock.title = "Test PR"
    return mock


@pytest.fixture
def mock_team() -> GithubTeam:
    mock = MagicMock(spec=GithubTeam)
    mock.organization.login = "Greenbax"
    mock.slug = "test-team"
    return mock


def test_create_slack_message(mock_pull_request: PullRequest) -> None:
    review_request = ReviewRequest(
        team="@Greenbax/test-team",
        channel="test-channel",
        slack_id="U123",
        pull_request=mock_pull_request,
    )
    message = create_slack_message(review_request)
    assert "U123" in message
    assert "PR-1" in message
    assert "Test PR" in message
    assert "test-team" in message


@patch(
    "bar_raiser.autofixes.notify_reviewer_teams.get_slack_user_icon_url_and_username"
)
@patch("bar_raiser.autofixes.notify_reviewer_teams.post_a_slack_message")
def test_process_review_request_success(
    mock_post_message: MagicMock,
    mock_get_user_info: MagicMock,
    mock_team: GithubTeam,
    mock_pull_request: PullRequest,
) -> None:
    mock_get_user_info.return_value = ("icon_url", "username")

    with patch(
        "pathlib.Path.read_text",
        return_value=dumps({"@Greenbax/test-team": "test-channel"}),
    ):
        comment, success = process_review_request(
            mock_team,
            mock_pull_request,
            "U123",
            dry_run="test-channel",
            github_team_to_slack_channels_path=Path("test-path"),
            github_team_to_slack_channels_help_msg="",
        )
    assert success
    assert "test-channel" in comment
    mock_post_message.assert_called_once()


@patch(
    "bar_raiser.autofixes.notify_reviewer_teams.get_slack_user_icon_url_and_username"
)
@patch("bar_raiser.autofixes.notify_reviewer_teams.post_a_slack_message")
def test_process_review_request_none_channel(
    mock_post_message: MagicMock,
    mock_get_user_info: MagicMock,
    mock_team: GithubTeam,
    mock_pull_request: PullRequest,
) -> None:
    mock_get_user_info.return_value = ("icon_url", "username")

    with patch(
        "pathlib.Path.read_text",
        return_value=dumps({}),
    ):
        comment, success = process_review_request(
            mock_team,
            mock_pull_request,
            "U123",
            dry_run="test-channel",
            github_team_to_slack_channels_path=Path("test-path"),
            github_team_to_slack_channels_help_msg="",
        )
    assert not success
    assert "Slack channel not found" in comment
    mock_post_message.assert_not_called()


@patch(
    "bar_raiser.autofixes.notify_reviewer_teams.get_slack_user_icon_url_and_username"
)
@patch("bar_raiser.autofixes.notify_reviewer_teams.post_a_slack_message")
def test_process_review_request_dry_run(
    mock_post_message: MagicMock,
    mock_get_user_info: MagicMock,
    mock_team: GithubTeam,
    mock_pull_request: PullRequest,
) -> None:
    mock_get_user_info.return_value = ("icon_url", "username")

    with patch(
        "pathlib.Path.read_text",
        return_value=dumps({"@Greenbax/test-team": "test-channel"}),
    ):
        comment, success = process_review_request(
            mock_team,
            mock_pull_request,
            "U123",
            dry_run="test-channel",
            github_team_to_slack_channels_path=Path("test-path"),
            github_team_to_slack_channels_help_msg="",
        )
    assert success
    assert "test-team" in comment
    mock_post_message.assert_called_once()


@patch("bar_raiser.autofixes.notify_reviewer_teams.get_slack_channel_from_mapping_path")
def test_process_pull_request_no_slack_id(
    mock_get_slack_id: MagicMock,
    mock_pull_request: PullRequest,
) -> None:
    mock_get_slack_id.return_value = None

    comment = process_pull_request(
        mock_pull_request,
        dry_run="test-channel",
        github_login_to_slack_ids_path=Path("test-path-1"),
        github_login_to_slack_ids_help_msg="Please update the mapping in `test-path-1`",
        github_team_to_slack_channels_path=Path("test-path-2"),
        github_team_to_slack_channels_help_msg="Please update the mapping in `test-path-2`",
        only_notify_team_slug=None,
    )
    assert (
        comment
        == "No author slack_id found for author testuser.\nPlease update the mapping in `test-path-1`\n"
    )


@patch("bar_raiser.autofixes.notify_reviewer_teams.get_slack_channel_from_mapping_path")
@patch("bar_raiser.autofixes.notify_reviewer_teams.process_review_request")
def test_process_pull_request_success(
    mock_process_request: MagicMock,
    mock_get_slack_id: MagicMock,
    mock_pull_request: PullRequest,
    mock_team: GithubTeam,
) -> None:
    mock_get_slack_id.return_value = "U123"
    mock_process_request.return_value = ("Test comment", True)

    mock_pull_request.get_review_requests = MagicMock(return_value=[[mock_team]])

    comment = process_pull_request(
        mock_pull_request,
        dry_run="test-channel",
        github_login_to_slack_ids_path=Path("test-path-1"),
        github_login_to_slack_ids_help_msg="",
        github_team_to_slack_channels_path=Path("test-path-2"),
        github_team_to_slack_channels_help_msg="",
        only_notify_team_slug=None,
    )
    assert comment == "Test comment"
    mock_process_request.assert_called_once()


@pytest.mark.parametrize(
    (
        "only_notify_team_slug_arg",
        "requested_teams_slugs",
        "process_review_request_return_value",
        "expected_comment",
        "expected_calls_to_process_review_request",
    ),
    [
        (
            "target-team",  # Test 1: Target team is a reviewer
            ["target-team", "other-team"],
            ("Comment for target-team", True),
            "Comment for target-team",
            1,
        ),
        (
            "non-existent-team",  # Test 2: Target team is NOT a reviewer
            ["team-a", "team-b"],
            ("Should not be called", True),
            "",
            0,
        ),
        (
            None,  # Test 3: only_notify_team_slug is None (all teams)
            ["team-1", "team-2"],
            ("Comment for team", True),
            "Comment for teamComment for team",  # Called for both teams
            2,
        ),
        (
            "target-team-no-channel",  # Test 4: Target team is reviewer, but process_review_request returns empty (e.g. no channel)
            ["target-team-no-channel", "another-team"],
            (
                "",
                False,
            ),  # Simulates an issue like a missing Slack channel mapping for the target team
            "",
            1,
        ),
        (
            "target-team",  # Test 5: Target team is a reviewer, no other reviewers
            ["target-team"],
            ("Comment for target-team", True),
            "Comment for target-team",
            1,
        ),
        (
            "not-a-reviewer",  # Test 6: Target team specified, but PR has no reviewers
            [],
            ("Should not be called", False),
            "",
            0,
        ),
        (
            None,  # Test 7: No target team, no reviewers
            [],
            ("Should not be called", False),
            "No team review requests found.",  # Updated expected message based on recent changes
            0,
        ),
    ],
)
@patch("bar_raiser.autofixes.notify_reviewer_teams.get_slack_channel_from_mapping_path")
@patch("bar_raiser.autofixes.notify_reviewer_teams.process_review_request")
def test_process_pull_request_only_notify_team(  # noqa: PLR0917
    mock_process_review_request: MagicMock,
    mock_get_author_slack_id: MagicMock,
    mock_pull_request: PullRequest,
    only_notify_team_slug_arg: str | None,
    requested_teams_slugs: list[str],
    process_review_request_return_value: tuple[str, bool],
    expected_comment: str,
    expected_calls_to_process_review_request: int,
) -> None:
    mock_get_author_slack_id.return_value = "U_AUTHOR123"  # For author lookup
    mock_process_review_request.return_value = process_review_request_return_value

    # Setup mock teams based on requested_teams_slugs
    mock_teams: list[MagicMock] = []
    for slug in requested_teams_slugs:
        team = MagicMock(spec=GithubTeam)
        team.organization.login = "Greenbax"
        team.slug = slug
        mock_teams.append(team)

    mock_pull_request.get_review_requests = MagicMock(return_value=[mock_teams])

    comment = process_pull_request(
        pull_request=mock_pull_request,
        dry_run="",  # Dry run channel not relevant for this focused test
        github_login_to_slack_ids_path=Path("dummy_login_to_slack.json"),
        github_login_to_slack_ids_help_msg="login help",
        github_team_to_slack_channels_path=Path("dummy_team_to_channels.json"),
        github_team_to_slack_channels_help_msg="team help",
        only_notify_team_slug=only_notify_team_slug_arg,
    )

    assert comment == expected_comment
    assert (
        mock_process_review_request.call_count
        == expected_calls_to_process_review_request
    )

    if expected_calls_to_process_review_request == 1 and only_notify_team_slug_arg:
        # Check that process_review_request was called with the correct target team
        called_team_arg = mock_process_review_request.call_args[0][0]
        assert called_team_arg.slug == only_notify_team_slug_arg


@patch("bar_raiser.autofixes.notify_reviewer_teams.get_pull_request")
def test_main(mock_get_pull_request: MagicMock) -> None:
    mock_get_pull_request.return_value = None
    with pytest.raises(SystemExit) as exc:
        main()
    assert exc.value.code == 2

    mock_pull_request = MagicMock(spec=PullRequest)
    mock_pull_request.labels = []
    mock_pull_request.draft = False
    mock_get_pull_request.return_value = mock_pull_request
    mock_label = MagicMock(spec=Label)
    mock_label.name = LABEL_TO_REMOVE
    mock_pull_request.labels = [mock_label]
    with (
        patch(
            "bar_raiser.autofixes.notify_reviewer_teams.process_pull_request",
            return_value="Test comment",
        ),
        patch(
            "sys.argv",
            [
                "notify_reviewer_teams.py",
                "github_login_to_slack_ids.json",
                "github_login_to_slack_ids_help_msg",
                "github_team_to_slack_channels.json",
                "github_team_to_slack_channels_help_msg",
            ],
        ),
    ):
        main()
    mock_pull_request.remove_from_labels.assert_called_once_with(LABEL_TO_REMOVE)



from __future__ import annotations

from datetime import datetime
from enum import StrEnum
from logging import INFO, basicConfig, getLogger
from os import environ
from pathlib import Path
from subprocess import check_output
from sys import stdout
from typing import TYPE_CHECKING, Literal, TypedDict
from zoneinfo import ZoneInfo

from git.repo import Repo
from github import Github, GithubIntegration, InputGitTreeElement
from github.CheckRun import CheckRun

if TYPE_CHECKING:
    from github.PullRequest import PullRequest
    from github.Repository import Repository

ANNOTATION_PAGE_SIZE = 50

logger = getLogger(__name__)


def get_github() -> Github:
    integration = GithubIntegration(environ["APP_ID"], environ["PRIVATE_KEY"])
    owner = environ["GITHUB_REPOSITORY_OWNER"]
    short_repo = environ["GITHUB_REPOSITORY"][len(owner) + 1 :]
    install = integration.get_installation(owner, short_repo)
    token = integration.get_access_token(install.id).token
    return Github(token)


def get_github_repo() -> Repository:
    github = get_github()
    logger.info(github.get_rate_limit())
    repo_name = environ["GITHUB_REPOSITORY"]
    return github.get_repo(repo_name)


def get_pull_request() -> PullRequest | None:
    try:
        return get_github_repo().get_pull(int(environ["PULL_NUMBER"]))
    except Exception:
        return None


def get_head_sha() -> str:
    if get_pull_request():  # pull requests have a virtual merge commit
        return get_git_repo().head.commit.parents[-1].hexsha
    return get_git_repo().head.commit.hexsha


def has_previous_issue_comment(pull: PullRequest, author: str, body: str) -> bool:
    for comment in pull.get_issue_comments():
        if comment.user.login == author and body in comment.body:
            return True
    return False


def get_git_repo() -> Repo:
    return Repo(".", search_parent_directories=True)


class Annotation(TypedDict):
    path: str
    start_line: int
    end_line: int
    annotation_level: str
    message: str


class Action(TypedDict):
    label: str
    description: str
    identifier: str


def create_check_run(
    *,
    repo: Repository,
    name: str,
    head_sha: str,
    conclusion: Literal["action_required", "success", "failure"],
    title: str,
    summary: str,
    annotations: list[Annotation],
    actions: list[Action],
) -> list[CheckRun]:
    checks = list[CheckRun]()
    while True:
        batch, annotations = (
            annotations[:ANNOTATION_PAGE_SIZE],
            annotations[ANNOTATION_PAGE_SIZE:],
        )
        check = repo.create_check_run(
            name=name,
            head_sha=head_sha,
            conclusion=conclusion,
            output={  # pyright: ignore[reportArgumentType]
                "title": title,
                "summary": summary,
                "annotations": batch,
            },
            actions=actions,  # pyright: ignore[reportArgumentType]
        )
        logger.info(check.html_url)
        checks.append(check)
        if len(annotations) == 0:
            break
    return checks


def commit_changes(
    repo: Repository, branch: str, sha: str, paths: list[str], commit_message: str
) -> None:
    batch_size = 200
    num_batches = len(paths) // batch_size + 1
    count = 0
    while paths:
        elements: list[InputGitTreeElement] = []
        for path in paths[:batch_size]:
            content = open(path, encoding="utf-8").read() if Path(path).exists() else ""
            blob = repo.create_git_blob(content, "utf-8")
            elements.append(InputGitTreeElement(path, "100644", "blob", sha=blob.sha))
        existing_tree = repo.get_git_tree(sha)
        existing_commit = repo.get_git_commit(sha)
        new_tree = repo.create_git_tree(elements, existing_tree)
        message = (
            f"{commit_message} ({count + 1}/{num_batches})"
            if num_batches > 1
            else commit_message
        )
        new_commit = repo.create_git_commit(message, new_tree, [existing_commit])
        repo.get_git_ref(f"heads/{branch}").edit(new_commit.sha)
        sha = new_commit.sha
        paths = paths[batch_size:]
        count += 1


def get_updated_paths(pull: PullRequest) -> list[str]:
    return [
        file.filename
        for file in pull.get_files()
        if file.status != "removed" and file.filename.endswith(".py")
    ]


def run_codemod_and_commit_changes(
    github_repo: Repository,
    pull_number: int,
    commands: list[list[str]],
    commit_message: str,
    run_on_updated_paths: bool,
) -> None:
    pull = github_repo.get_pull(pull_number)
    if run_on_updated_paths:
        pull_updated_paths = get_updated_paths(pull)
        commands = [command + pull_updated_paths for command in commands]

    for command in commands:
        check_output(command)

    git_repo = get_git_repo()
    updated_paths: list[str] = [
        i.b_path  # pyright: ignore[reportUnknownMemberType]
        for i in git_repo.index.diff(  # pyright: ignore[reportUnknownMemberType,reportUnknownVariableType]
            None
        )
    ]
    if run_on_updated_paths:
        updated_paths = [path for path in updated_paths if path in updated_paths]

    logger.info(f"Updated paths: {updated_paths}")
    if len(updated_paths) == 0:
        msg = "No updated paths to commit."
        pull.create_issue_comment(msg)
        logger.info(msg)
    else:
        commit_changes(
            repo=github_repo,
            branch=pull.head.ref,
            sha=pull.head.sha,
            paths=updated_paths,
            commit_message=commit_message,
        )
        logger.info(f"Changes committed to {pull.html_url}.")


def create_a_pull_request(
    github_repo: Repository,
    codemod: str,
    actor: str,
    extra_body: str = "",
) -> None:
    git_repo = get_git_repo()
    codemod_name = (
        codemod.strip().split("codemod ")[1].split(" ")[0]
        if "codemod " in codemod
        else codemod.replace(" ", "-")
    )
    updated_paths: list[str] = [
        i.b_path  # pyright: ignore[reportUnknownMemberType]
        for i in git_repo.index.diff(  # pyright: ignore[reportUnknownMemberType,reportUnknownVariableType]
            None
        )
    ]
    if len(updated_paths) == 0:
        msg = "No updated paths to commit."
        logger.info(msg)
    else:
        logger.info(f"Updated paths: {updated_paths}")
        commit = github_repo.get_commit("master")
        time = datetime.now(ZoneInfo("US/Pacific")).strftime("%Y%m%d-%H-%M-%S")
        branch = f"codemod-{codemod_name}-{actor}-{time}"
        ref = f"refs/heads/{branch}"
        github_repo.create_git_ref(ref, commit.sha)
        title = f"Codemod {codemod_name} by {actor} at {time}"
        commit_changes(
            repo=github_repo,
            branch=branch,
            sha=commit.sha,
            paths=updated_paths,
            commit_message=title,
        )
        body = f"Codemod Command: `{codemod}`\nCreated from {environ['GITHUB_SERVER_URL']}/{environ['GITHUB_REPOSITORY']}/actions/runs/{environ['GITHUB_RUN_ID']} by {actor}"
        if extra_body:
            body += f"\n\n{extra_body}"
        pull = github_repo.create_pull(title=title, body=body, base="master", head=ref)
        pull.add_to_assignees(actor)
        pull.enable_automerge(merge_method="SQUASH")
        logger.info(f"Changes committed to {pull.html_url}")


def initialize_logging() -> None:
    basicConfig(
        stream=stdout,
        level=INFO,
        format="%(asctime)s %(levelname)s %(module)s - %(funcName)s: %(message)s",
    )


class Autofixes(StrEnum):
    RUFF = "autofix-ruff"
    PYRIGHT_IGNORES = "pyright-ignores"



from __future__ import annotations

import os
import re
import subprocess
import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar, cast

from libcst import Comment, Module, RemovalSentinel
from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand
from libcst.metadata import PositionProvider

if TYPE_CHECKING:
    from libcst.metadata import CodeRange


@dataclass(frozen=True, slots=True)
class PyrightError:
    rule: str


class RemoveUnnecessaryPyrightIgnoreComments(VisitorBasedCodemodCommand):
    DESCRIPTION = """
    Removes unnecessary `pyright: ignore` comments.

    Command: ```

    python -m libcst.tool codemod --jobs=1 remove_unnecessary_pyright_ignore_comments.RemoveUnnecessaryPyrightIgnoreComments <path/to/file.py>

    ```
    """

    COMMENT_REGEX = r"(?:pyright|type): ignore(?:\[(.+)\])?"

    COMMENT_RULE_REGEX = r"pyright: ignore\[([a-zA-Z,]+)\]"

    COMMENT_VALUE = "pyright: ignore"

    METADATA_DEPENDENCIES = (PositionProvider,)

    # {filename}:{line}:{column} - error: Unnecessary " " rule: "{rule}"
    PYRIGHT_ERROR_REGEX = (
        r'(.+):(\d+):(\d+) - error: Unnecessary "# pyright: ignore" rule: "(\w+)"'
    )

    pyright_errors_by_comment: ClassVar[dict[Comment, list[PyrightError]]] = {}

    pyright_errors_by_line: ClassVar[dict[int, list[PyrightError]]] = {}

    pyright_errors_by_line_by_filename: ClassVar[
        dict[str, dict[int, list[PyrightError]]]
    ] = {}

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)

        filenames = [self.context.filename] if self.context.filename else sys.argv[3:]
        pyright_project = cast(
            "str | None", self.context.scratch.get("pyright_project")
        )

        pyright_args = (
            [
                "pyright",
                "-p",
                pyright_project,
                *filenames,
            ]
            if pyright_project
            else ["pyright", *filenames]
        )

        pyright_stdout = (
            os.getenv("PYRIGHT_OUTPUT")
            or subprocess.run(
                pyright_args,
                capture_output=True,
                check=False,
                env=dict(os.environ, NODE_OPTIONS="--max-old-space-size=8192"),
                text=True,
            ).stdout
        )

        self._set_pyright_errors_by_line_by_filename(pyright_stdout)

    def _set_pyright_errors_by_line_by_filename(self, pyright_stdout: str) -> None:
        for filename, line, *_, rule in re.findall(
            self.PYRIGHT_ERROR_REGEX, pyright_stdout
        ):
            self.pyright_errors_by_line_by_filename.setdefault(
                filename.strip(), {}
            ).setdefault(int(line), []).append(PyrightError(rule=rule))

    def visit_Module(self, node: Module) -> bool | None:
        pyright_stdout = os.getenv("PYRIGHT_OUTPUT")

        if pyright_stdout:
            self._set_pyright_errors_by_line_by_filename(pyright_stdout)

        self.pyright_errors_by_line = {}  # pyright: ignore[reportAttributeAccessIssue]

        if (
            not self.context.filename
            or self.context.filename not in self.pyright_errors_by_line_by_filename
        ):
            return False

        self.pyright_errors_by_line = self.pyright_errors_by_line_by_filename[  # pyright: ignore[reportAttributeAccessIssue]
            self.context.filename
        ]

        return True

    def visit_Comment(self, node: Comment) -> bool | None:
        metadata = cast("CodeRange", self.get_metadata(PositionProvider, node))

        if metadata.start.line in self.pyright_errors_by_line:
            self.pyright_errors_by_comment[node] = self.pyright_errors_by_line[
                metadata.start.line
            ]

        return True

    def leave_Comment(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, original_node: Comment, updated_node: Comment
    ) -> Comment | RemovalSentinel:
        if original_node not in self.pyright_errors_by_comment:
            return updated_node

        comment_rules = {
            comment_rule
            for _comment_rules in re.findall(
                self.COMMENT_RULE_REGEX, (original_node and original_node.value) or ""
            )
            for comment_rule in _comment_rules.split(",")
            if _comment_rules and comment_rule
        }

        comment_value: str | None = None

        pyright_errors = self.pyright_errors_by_comment[original_node]

        rules = comment_rules - {pyright_error.rule for pyright_error in pyright_errors}

        if rules:
            rules_str = ",".join(sorted(rules))

            pyright_ignore_comment_value = f"{self.COMMENT_VALUE}[{rules_str}]"

            comment_value = " ".join([
                pyright_ignore_comment_value,
                re.sub(self.COMMENT_REGEX, "", original_node.value)
                .replace("#", "")
                .strip(),
            ]).strip()

        return (
            Comment(f"# {comment_value}") if comment_value else RemovalSentinel.REMOVE
        )



from __future__ import annotations

import json
import operator
import os
from calendar import monthrange
from collections import Counter, defaultdict
from dataclasses import asdict, dataclass, is_dataclass
from datetime import datetime
from enum import StrEnum
from importlib import import_module
from inspect import isclass
from logging import getLogger
from pathlib import Path
from pkgutil import walk_packages
from re import match
from typing import TYPE_CHECKING, Any, cast

import libcst as cst
from dateutil.relativedelta import relativedelta
from fixit import CstLintRule
from fixit.cli import (
    map_paths,  # pyright: ignore[reportUnknownVariableType]
)
from fixit.cli.args import LintWorkers
from fixit.common.base import CstContext, LintConfig
from fixit.common.config import get_lint_config
from fixit.rule_lint_engine import _visit_cst_rules_with_context
from libcst.metadata import CodePosition, MetadataWrapper, PositionProvider

if TYPE_CHECKING:
    from collections.abc import Collection, Iterable, Iterator, Mapping

    from fixit.common.report import BaseLintRuleReport
    from git import Commit, DiffIndex
    from git.diff import Diff
    from git.repo import Repo
    from libcst.metadata.base_provider import ProviderT
    from types_aiobotocore_s3 import S3Client


logger = getLogger(__name__)

S3_BUCKET = "s3-bucket"
S3_KEY_LEADERBOARD = "bar-raiser/leaderboard/"
S3_KEY_PATH_RESULTS = "bar-raiser/path_results/"
S3_KEY_HISTORY = "bar-raiser/history/"


@dataclass(frozen=True)
class Result:
    key: str
    path: str
    line: int
    col: int
    code: str
    patch: None
    count: int = 1


class CodeAnalyzerContext(CstContext):
    def __init__(
        self,
        wrapper: MetadataWrapper,
        source: bytes,
        file_path: Path,
        config: LintConfig,
    ) -> None:
        self.results: list[Result] = []
        super().__init__(wrapper, source, file_path, config)


class BaseCodeAnalyzer(CstLintRule):
    context: CodeAnalyzerContext

    def report(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        node: cst.CSTNode,
        key: str,
        message: str | None = None,
        *,
        count: int = 1,
        position: CodePosition | None = None,
        replacement: cst.CSTNode | cst.RemovalSentinel | None = None,
    ) -> None:
        if position is None:
            position = self.context.wrapper.resolve(PositionProvider)[node].start
        self.context.results.append(
            Result(
                key=key,
                path=str(self.context.file_path),
                line=position.line,
                col=position.column,
                code=self.__class__.__name__,
                count=count,
                patch=None,
            )
        )
        return super().report(  # pyright: ignore[reportUnknownMemberType]
            node, message, position=position, replacement=replacement
        )


def lint_file(
    file_path: Path,
    source: bytes,
    *,
    use_ignore_byte_markers: bool = True,
    use_ignore_comments: bool = True,
    config: LintConfig,
    rules: Collection[type[CstLintRule]] = (),
    cst_wrapper: MetadataWrapper | None = None,
    find_unused_suppressions: bool = False,
) -> list[Result] | list[BaseLintRuleReport]:
    if cst_wrapper is None:
        cst_wrapper = MetadataWrapper(cst.parse_module(source), unsafe_skip_copy=True)

    use_code_analyzer = any(issubclass(rule, BaseCodeAnalyzer) for rule in rules)
    if use_code_analyzer:
        context = CodeAnalyzerContext(cst_wrapper, source, file_path, config)
        _visit_cst_rules_with_context(cst_wrapper, rules, context)
        return context.results
    context = CstContext(cst_wrapper, source, file_path, config)
    _visit_cst_rules_with_context(cst_wrapper, rules, context)
    return context.reports


@dataclass(frozen=True)
class LintOptions:
    rules: Collection[type[CstLintRule]]
    config: LintConfig


def index_to_coordinates(text: str, index: int) -> tuple[int, int]:
    """Returns (line_number, col) of `index` in `s`."""
    if not len(text):
        return 1, 1
    sp = text[: index + 1].splitlines(keepends=True)
    return len(sp), len(sp[-1])


class TechDebtCategory(StrEnum):
    WEIGHTED_SCORE = "Weighted Score"
    PYRIGHT_IGNORE = "pyright-ignore"
    BE_LINES = "BE-lines"


REGRESSION_TECH_DEBT_CATEGORIES = [  # Tech debt categories that could regress over time
    TechDebtCategory.BE_LINES,  # Only earn scores when making net reduction and will not lose scores when adding more debts
]

TECH_DEBT_SHOUT_OUT_THRESHOLD: dict[TechDebtCategory, int] = {
    TechDebtCategory.PYRIGHT_IGNORE: -1,
    TechDebtCategory.BE_LINES: -25,
}

TECH_DEBT_CONTRIBUTION_EXPLANATION: dict[TechDebtCategory, str] = {
    TechDebtCategory.PYRIGHT_IGNORE: "that improved backend type safety",
    TechDebtCategory.BE_LINES: "that cleaned up unused backend code and improved code quality",
}

TECH_DEBT_IMPACT: dict[TechDebtCategory, str] = {
    TechDebtCategory.PYRIGHT_IGNORE: "Fixing Pyright ignore comments can enhance type safety and reduce runtime errors by ensuring that all type issues are addressed during development.",
    TechDebtCategory.BE_LINES: "Removing unused backend code can simplify the codebase, making it easier to maintain and reducing potential confusion.",
}

CURRNET_QUARTER = "Q2"
QUARTER_BEGINNING = datetime(2025, 4, 1)
quarter_end = QUARTER_BEGINNING + relativedelta(months=2)
last_day = monthrange(quarter_end.year, quarter_end.month)[1]
QUARTER_END = quarter_end.replace(day=last_day)

TECH_DEBT_BEGINNING_AND_GOALS: dict[TechDebtCategory, tuple[int, int]] = {}

TECH_DEBT_WEIGHTS: dict[TechDebtCategory, int] = {
    TechDebtCategory.PYRIGHT_IGNORE: 10,
    TechDebtCategory.BE_LINES: 1,
}


TROPHY_EMOJI = ":trophy:"

NEW_TECH_DEBT_MESSAGE = "You're about to introduce new tech debt. Would you be able to address it upfront or, alternatively, tackle some existing tech debt from http://go/quality-wins offset the impact? :pray:"


def find_files(paths: Iterable[str | Path]) -> Iterator[str]:
    """
    Given an iterable of paths, yields any files and walks over any directories.
    """
    for path in paths:
        if os.path.isfile(path):
            yield str(path)
        else:
            for root, _dirs, files in os.walk(path):
                for f in files:
                    if not os.path.islink(f):
                        yield os.path.join(root, f)


def get_analyzed_results(
    path: Path, options: LintOptions, _: Mapping[ProviderT, object] | None
) -> list[Result]:
    try:
        if ".venv/" in str(path):
            return []
        source = path.read_bytes()
        source_str = source.decode("utf-8")
        match path.suffix:
            case ".py":
                results: list[Result] = cast(
                    "list[Result]",
                    lint_file(
                        path,
                        source,
                        rules=options.rules,
                        config=options.config,
                        cst_wrapper=None,
                        find_unused_suppressions=True,
                    ),
                )
                results.append(
                    Result(
                        key=TechDebtCategory.BE_LINES.value,
                        path=str(path),
                        line=1,
                        col=1,
                        code=TechDebtCategory.BE_LINES.value,
                        patch=None,
                        count=sum(1 for _ in source_str.split("\n")),
                    )
                )
                return results
    except Exception as exp:
        print(exp)
    return []


class DataclassJSONEncoder(json.JSONEncoder):
    def default(
        self,
        o: Any,
    ):
        if is_dataclass(o):
            return asdict(o)  # pyright: ignore[reportArgumentType]
        return super().default(o)


def get_analyzers() -> set[type[BaseCodeAnalyzer]]:
    package = import_module("bar_raiser.tech_debt_framework.analyzers")
    analyzers: set[type[BaseCodeAnalyzer]] = set()
    for _loader, name, _is_pkg in walk_packages(package.__path__):
        full_name = package.__name__ + "." + name
        try:
            module = import_module(full_name)
            for obj_name in dir(module):
                obj = getattr(module, obj_name)
                if (
                    isclass(obj)
                    and obj is not BaseCodeAnalyzer
                    and issubclass(obj, BaseCodeAnalyzer)
                ):
                    analyzers.add(obj)
        except ModuleNotFoundError:
            pass
    return analyzers


def remove_delta(
    delta: Counter[TechDebtCategory], path_results: PathResults, path: str
) -> None:
    if path and path in path_results:
        for result in path_results[path]:
            delta[TechDebtCategory(result.key)] -= result.count
        del path_results[path]


def get_delta(
    path_results: PathResults,
    diffs: DiffIndex[Diff],
    analyzers: set[type[BaseCodeAnalyzer]],
) -> Counter[TechDebtCategory]:
    delta = Counter[TechDebtCategory]()
    updated_paths: set[str] = set()
    for diff in diffs:
        if diff.a_path:
            remove_delta(delta, path_results, diff.a_path)
        b_path = diff.b_path
        if b_path:
            updated_paths.add(b_path)

    results = PathResults.analyze_paths(list(updated_paths), analyzers)
    for path, rlts in results.items():
        for result in rlts:
            delta[TechDebtCategory(result.key)] += result.count
        path_results[path] = rlts
    zero_keys = [key for key, val in delta.items() if val == 0]
    for key in zero_keys:
        del delta[key]
    return delta


def get_delta_value_with_emoji(delta_value: int, with_emoji: bool = True) -> str:
    if delta_value > 0:
        return f"{delta_value:+,d}{' :cry:' if with_emoji else ''}"
    if delta_value < 0:
        return f"{delta_value:,} :tada:"
    return str(f"{delta_value:,}")


def get_progress_text(beginning: int, goal: int, current: int) -> str:
    today = datetime.now()
    expected = beginning + (today - QUARTER_BEGINNING).total_seconds() / (
        QUARTER_END - QUARTER_BEGINNING
    ).total_seconds() * (goal - beginning)
    if current > expected:
        return f":warning: expecting {expected:,.0f}"
    return ":white_check_mark: on track"


class PathResults(dict[str, list[Result]]):
    @staticmethod
    def load(path: str) -> PathResults:
        path_results = PathResults()
        with open(path, encoding="utf-8") as f:
            raw_path_results = json.load(f)
            for sub_path, results in raw_path_results.items():
                path_results[sub_path] = [Result(**result) for result in results]
        return path_results

    @staticmethod
    async def load_with_commit(s3: S3Client, commit: Commit) -> PathResults:
        local_json_path = f"path_results-{commit.hexsha}.json"
        if not Path(local_json_path).exists():
            s3_key = f"{S3_KEY_PATH_RESULTS}{commit.hexsha}.json"
            await s3.download_file(
                S3_BUCKET,
                s3_key,
                local_json_path,
            )
            logger.info(f"Successfully downloaded {s3_key}")
        return PathResults.load(local_json_path)

    def dump(self, path: str) -> None:
        json.dump(
            obj=self,
            fp=open(path, "w", encoding="utf-8"),
            cls=DataclassJSONEncoder,
            indent=2,
        )

    async def upload_to_s3(self, s3: S3Client, commit: Commit) -> None:
        local_json_path = f"path_results-{commit.hexsha}.json"
        self.dump(local_json_path)
        s3_key = f"{S3_KEY_PATH_RESULTS}{commit.hexsha}.json"
        await s3.upload_file(local_json_path, S3_BUCKET, s3_key)
        logger.info(f"Successfully uploaded {s3_key}")

    def __getitem__(self, __key: str) -> list[Result]:
        if __key not in self:
            super().__setitem__(__key, [])
        return super().__getitem__(__key)

    @staticmethod
    def analyze_paths(
        paths: list[str],
        analyzers: set[type[BaseCodeAnalyzer]],
        workers: LintWorkers = LintWorkers.CPU_COUNT,
    ) -> PathResults:
        path_results = PathResults()
        print(analyzers)
        for results in map_paths(  # pyright: ignore[reportUnknownVariableType]
            get_analyzed_results,
            {str(path) for path in find_files(paths)},
            LintOptions(
                rules=analyzers,
                config=get_lint_config(),
            ),
            workers=workers,
        ):
            for result in results:  # pyright: ignore[reportUnknownVariableType]
                result = cast("Result", result)
                path_results[result.path].append(result)
        return path_results

    def get_key_counts(self) -> Counter[TechDebtCategory]:
        counter = Counter[TechDebtCategory]()
        for results in self.values():
            for result in results:
                counter[TechDebtCategory(result.key)] += result.count
        return counter

    @staticmethod
    async def gen_from_incremental_analysis(  # noqa: PLR0917
        s3: S3Client,
        git_repo: Repo,
        base_commit: Commit,
        head_commit: Commit,
        paths: list[str],
        analyzers: set[type[BaseCodeAnalyzer]],
        force_recompute: bool = False,
    ) -> tuple[Counter[TechDebtCategory], PathResults]:
        if force_recompute:
            git_repo.git.checkout(base_commit)
            logger.info(f"git checkout {base_commit}")
            path_results = PathResults.analyze_paths(paths, analyzers)
            git_repo.git.checkout(head_commit)
            logger.info(f"git checkout {head_commit}")
        else:
            try:
                path_results = await PathResults.load_with_commit(s3, base_commit)
            except Exception:
                logger.warning(
                    f"Failed to download PathResults. Checkout {base_commit}"
                )
                git_repo.git.checkout(base_commit)
                path_results = PathResults.analyze_paths(paths, analyzers)
                try:
                    await path_results.upload_to_s3(s3, base_commit)
                except Exception:
                    logger.warning("Failed to upload PathResults to s3.")
                git_repo.git.checkout(head_commit)

        delta = get_delta(
            path_results,
            base_commit.diff(  # pyright: ignore[reportUnknownArgumentType,reportUnknownMemberType]
                head_commit
            ),
            analyzers,
        )
        logger.info(f"Delta: {delta}")
        try:
            await path_results.upload_to_s3(s3, head_commit)
        except Exception:
            logger.warning("Failed to upload PathResults to s3.")
        return delta, path_results

    def get_markdown_summary(
        self,
        delta: Counter[TechDebtCategory],
        author: str,
        cumulative_delta: Counter[TechDebtCategory] | None = None,
    ) -> str:
        overall = self.get_key_counts()

        show_regression_notice = False
        for key, value in delta.items():
            if key not in REGRESSION_TECH_DEBT_CATEGORIES and value > 0:
                show_regression_notice = True

        output = f"## Tech Debt Contribution Summary for {author}\n\n"
        if show_regression_notice:
            output += f"> [!CAUTION]\n> {NEW_TECH_DEBT_MESSAGE}\n\n"
        output += f"| Tech Debt | {CURRNET_QUARTER} Goal | Remaining | Progress Check | Weight | Upcoming Contribution from this PR |"
        if cumulative_delta:
            output += f" Your Contribution in 2025 {CURRNET_QUARTER} |"
        output += "\n|:---------:|:----------:|:----------:|:----------:|:----------:|:-----------:|"
        if cumulative_delta:
            output += ":-----------:|"
        output += "\n"

        for key, val in sorted(overall.items(), key=operator.itemgetter(0)):
            if key in TECH_DEBT_BEGINNING_AND_GOALS:
                beginning, goal = TECH_DEBT_BEGINNING_AND_GOALS[key]
                progress = get_progress_text(beginning, goal, val)
                goal_text = f"{goal:,}"
            else:
                progress = "N/A"
                goal_text = "N/A"
            output += (
                "| "
                + key
                + " | "
                + goal_text
                + " | "
                + f"{val:,}"
                + " | "
                + progress
                + " | "
                + str(TECH_DEBT_WEIGHTS.get(key, "N/A"))
                + " | "
                + get_delta_value_with_emoji(
                    delta.get(key, 0),
                    key not in REGRESSION_TECH_DEBT_CATEGORIES,
                )
            )
            if cumulative_delta:
                cumulative_value = cumulative_delta.get(key, 0)
                output += " | " + str(
                    get_delta_value_with_emoji(
                        cumulative_value,
                        key not in REGRESSION_TECH_DEBT_CATEGORIES,
                    )
                )

            output += " |\n"
        output += "\n"
        return output


class History:
    def __init__(self) -> None:
        self.data: dict[str, list[tuple[str, Counter[TechDebtCategory]]]] = {}

    @staticmethod
    def load(path: str) -> History:
        history = History()
        with open(path, encoding="utf-8") as f:
            raw_history = json.load(f)
            for author, items in raw_history.items():
                history.data[author] = [(sha, Counter(delta)) for sha, delta in items]
        return history

    @staticmethod
    async def load_with_commit(s3: S3Client, commit: Commit) -> History:
        local_json_path = f"history-{commit.hexsha}.json"
        if not Path(local_json_path).exists():
            s3_key = f"{S3_KEY_HISTORY}{commit.hexsha}.json"
            await s3.download_file(
                S3_BUCKET,
                s3_key,
                local_json_path,
            )
            logger.info(f"Successfully downloaded {s3_key}")
        return History.load(local_json_path)

    def dump(self, path: str) -> None:
        json.dump(
            obj=self.data,
            fp=open(path, "w", encoding="utf-8"),
            indent=2,
        )

    async def upload_to_s3(self, s3: S3Client, commit: Commit) -> None:
        local_json_path = f"history-{commit.hexsha}.json"
        self.dump(local_json_path)
        s3_key = f"{S3_KEY_HISTORY}{commit.hexsha}.json"
        await s3.upload_file(local_json_path, S3_BUCKET, s3_key)
        logger.info(f"Successfully uploaded {s3_key}")

    def add_delta(
        self,
        author: str,
        sha: str,
        delta: Counter[TechDebtCategory],
        weighted_score: int,
    ) -> None:
        if TechDebtCategory.WEIGHTED_SCORE not in delta:
            delta[TechDebtCategory.WEIGHTED_SCORE] = weighted_score
        if author not in self.data:
            self.data[author] = []
        self.data[author].append((sha, delta))

    def get_markdown_summary(self, author: str) -> str:
        output = f"## Tech Debt Contribution History for {author}\n\n"
        output += "> [!NOTE]\n> The scores in parentheses represent the weighted delta, which is added to the overall weighted score. The scores inside parenthes are weighted delta that is added to the weigted scores. There's no penalty for adding new FE-lines or BE-lines. You'll earn points in these categories when you achieve a net reduction in lines of code.\n\n"
        output += "| Commit | "
        if author not in self.data:
            return output
        tech_debt_keys = list(TechDebtCategory)
        for key in tech_debt_keys:
            output += f" {key} |"
        output += "\n|:----------|"
        for _ in tech_debt_keys:
            output += "----------:|"
        output += "\n"
        for sha, delta in self.data[author]:
            output += (
                f"| [{sha[:7]}](https://github.com/ZipHQ/bar-raisercommit/{sha}) |"
            )
            for key in tech_debt_keys:
                raw_score = delta.get(key, 0)
                output += get_delta_value_with_emoji(
                    raw_score,
                    key
                    not in {
                        *REGRESSION_TECH_DEBT_CATEGORIES,
                        TechDebtCategory.WEIGHTED_SCORE,
                    },
                )
                if raw_score != 0 and key not in {
                    *REGRESSION_TECH_DEBT_CATEGORIES,
                    TechDebtCategory.WEIGHTED_SCORE,
                }:
                    output += f" ({raw_score * TECH_DEBT_WEIGHTS[key]:+,d})"
                output += " |"
            output += "\n"
        return output


def ordinal(n: int):
    # Convert an integer to its ordinal representation.
    if 11 <= n % 100 <= 13:  # Special cases for 11th, 12th, 13th
        suffix = "th"
    else:
        suffix = {1: "st", 2: "nd", 3: "rd"}.get(n % 10, "th")
    return f"{n}{suffix}"


class LeaderBoard:
    contribution_summary: str = ""

    def __init__(self) -> None:
        self.board: dict[str, Counter[TechDebtCategory]] = defaultdict(Counter)

    @staticmethod
    def load(path: str) -> LeaderBoard:
        leaderboard = LeaderBoard()
        with open(path, encoding="utf-8") as f:
            raw_leaderboard = json.load(f)
            for author, delta in raw_leaderboard.items():
                leaderboard.board[author] = Counter(delta)
        return leaderboard

    @staticmethod
    async def load_with_commit(s3: S3Client, commit: Commit) -> LeaderBoard:
        local_json_path = f"leaderboard-{commit.hexsha}.json"
        if not Path(local_json_path).exists():
            s3_key = f"{S3_KEY_LEADERBOARD}{commit.hexsha}.json"
            await s3.download_file(
                S3_BUCKET,
                s3_key,
                local_json_path,
            )
            logger.info(f"Successfully downloaded {s3_key}")
        return LeaderBoard.load(local_json_path)

    def dump(self, path: str) -> None:
        json.dump(
            obj=self.board,
            fp=open(path, "w", encoding="utf-8"),
            indent=2,
        )

    async def upload_to_s3(self, s3: S3Client, commit: Commit) -> None:
        local_json_path = f"leaderboard-{commit.hexsha}.json"
        self.dump(local_json_path)
        s3_key = f"{S3_KEY_LEADERBOARD}{commit.hexsha}.json"
        await s3.upload_file(local_json_path, S3_BUCKET, s3_key)
        logger.info(f"Successfully uploaded {s3_key}")

    def add_delta_and_check_contribution(
        self, author: str, delta: Counter[TechDebtCategory]
    ) -> None:
        tech_debt_regression = ""
        for key, count in delta.items():
            if key in {
                *REGRESSION_TECH_DEBT_CATEGORIES,
                TechDebtCategory.WEIGHTED_SCORE,
            }:
                continue
            if count > 0:
                tech_debt_regression += f"- {key}: +{count}\n"
        self.tech_debt_regression = tech_debt_regression

        contribution_summary = ""
        for key, count in delta.items():
            if count <= TECH_DEBT_SHOUT_OUT_THRESHOLD.get(
                key, -1
            ):  # shout out when making a contribution meets the threshold
                if len(contribution_summary) > 0:
                    contribution_summary += ", and "
                contribution_summary += f"fixing *{count * -1} {key}* {TECH_DEBT_CONTRIBUTION_EXPLANATION.get(key, '')}"
            self.board[author][key] += count
        self.contribution_summary = contribution_summary

    def compute_weighted_score(self) -> None:
        for winner, delta in self.board.items():
            weighted_scores = 0
            for key, score in delta.items():
                if (
                    key in REGRESSION_TECH_DEBT_CATEGORIES and score > 0
                ) or key == TechDebtCategory.WEIGHTED_SCORE:
                    continue
                weighted_scores += score * TECH_DEBT_WEIGHTS[key]
            self.board[winner][TechDebtCategory.WEIGHTED_SCORE] = weighted_scores

    def get_markdown_summary(self, author: str, limit: int = 30) -> str:  # noqa: PLR0912
        tech_debt_keys = list(TechDebtCategory)
        output = f"## 2025 {CURRNET_QUARTER} Code Quality Award Leaderboard\n\n"
        output += f"> [!NOTE]\n> At the end of {CURRNET_QUARTER} ({QUARTER_END.strftime('%m/%d/%Y')}), an award committee will select 5 winners base on the [weighted](https://github.com/Greenbax/evergreen/blob/master/website/fixit_linters/analyzers/utils.py) scores to distribute the awards!\n\n"
        tech_debt_ranks: dict[str, list[tuple[str, int]]] = defaultdict(list)
        output += "|"
        for key in tech_debt_keys:
            output += f" {key} contribution | Rank |"
            for winner, delta in self.board.items():
                tech_debt_ranks[key].append((winner, delta[key]))
        output += "\n|"
        for _ in tech_debt_keys:
            output += "----------:|:----------|"
        output += "\n"

        for ranks in tech_debt_ranks.values():
            ranks.sort(key=operator.itemgetter(1))
        notes: dict[str, list[str]] = defaultdict(list)
        detail = ""
        num_trophies = 0
        for rank in range(len(self.board)):
            if rank < limit:
                detail += "|"
            for key in tech_debt_keys:
                winner, score = tech_debt_ranks[key][rank]
                winner_text = winner
                if winner == author:
                    winner_text = f"**{winner}** :point_left:"
                    notes[key].append(f"Your rank: **{ordinal(rank + 1)}**.")
                    if rank - 1 >= 0:
                        score_diff = tech_debt_ranks[key][rank - 1][1] - score
                        if score_diff == 0:
                            score_diff = -1
                        notes[key].append(
                            f"You need {score_diff:,} contribution to Rank Up :rocket:"
                        )
                        if (
                            key == TechDebtCategory.WEIGHTED_SCORE
                            and self.contribution_summary
                        ):
                            self.contribution_summary += f". You're currently ranked *{ordinal(rank + 1)}* and need *{-1 * score_diff:,}* more weighted scores to rank up 🚀"
                    elif (
                        key == TechDebtCategory.WEIGHTED_SCORE
                        and self.contribution_summary
                    ):
                        self.contribution_summary += f". You're currently ranked *{ordinal(rank + 1)}* 🚀 Keep it up!"
                spaces = 25 - len(winner_text)
                WINNER_EXCLUSIONS = set[str]([
                    "zip-bar-raiser[bot]",
                ])
                if rank < limit:
                    emoji = (
                        TROPHY_EMOJI
                        if num_trophies < 5
                        and key == TechDebtCategory.WEIGHTED_SCORE
                        and winner not in WINNER_EXCLUSIONS
                        else ""
                    )
                    if emoji == TROPHY_EMOJI:
                        num_trophies += 1
                    detail += f" {score:,} | {emoji} {rank + 1} {winner_text} {' ' * (max(0, spaces))} |"
            detail += "\n"
        if notes:
            output += "|"
            for key in tech_debt_keys:
                notes_list = notes[key]
                if notes_list:
                    output += f" {notes_list[0]} |"
                output += f" {notes_list[1]} |" if len(notes_list) > 1 else " |"
            output += "\n"
        output += detail
        return output


def get_pr_num_from_commit_message(commit_message: str) -> str:
    matched = match(
        r"^.*\(#(?P<PR_NUMBER>[0-9]+)\)$",
        commit_message,
    )
    return matched.group("PR_NUMBER") if matched else ""



from __future__ import annotations

import asyncio
from argparse import ArgumentParser
from collections import Counter
from logging import getLogger
from os import environ
from typing import TYPE_CHECKING

from aioboto3 import Session
from botocore.exceptions import ClientError
from git.repo import Repo

from bar_raiser.tech_debt_framework.utils import (
    NEW_TECH_DEBT_MESSAGE,
    REGRESSION_TECH_DEBT_CATEGORIES,
    BaseCodeAnalyzer,
    History,
    LeaderBoard,
    PathResults,
    TechDebtCategory,
    get_analyzers,
    get_pr_num_from_commit_message,
)
from bar_raiser.utils.github import (
    create_check_run,
    get_git_repo,
    get_github_repo,
    get_pull_request,
    has_previous_issue_comment,
    initialize_logging,
)
from bar_raiser.utils.slack import post_a_slack_message

if TYPE_CHECKING:
    from git import Commit
    from github.Repository import Repository
    from types_aiobotocore_s3 import S3Client

logger = getLogger(__name__)

CHECK_NAME = "quality-wins-report"
MAX_COMMENT_LENGTH = 65536
RAISE_THE_BAR_LOGIN = "zip-bar-raiser[bot]"
TECH_DEBT_SHOUTOUT_CHANNEL = "C0600000000"


def get_slack_handle_from_github_login(github_login: str) -> str:
    return ""


def trim_to_max_bytes(text: str, max_bytes: int) -> str:
    """
    Trim a string to ensure it's at most max_bytes in UTF-8 encoding.
    If trimming results in a partial character at the end, that character is removed.
    """
    encoded = text.encode("utf-8")
    if len(encoded) <= max_bytes:
        return text

    trimmed = encoded[:max_bytes]

    try:
        return trimmed.decode("utf-8")
    except UnicodeDecodeError:
        while trimmed:
            trimmed = trimmed[:-1]
            try:
                return trimmed.decode("utf-8")
            except UnicodeDecodeError:
                continue

    return ""  # Fallback, should not reach here


def get_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument("paths", nargs="*", help="List of paths to run lint rules on.")
    parser.add_argument(
        "--backfill",
        type=int,
        default=None,
        help="number of recent commits to backfill",
    )
    parser.add_argument(
        "--start-hexsha",
        type=str,
        help="A starting commit hexsha to backfill from",
    )
    parser.add_argument(
        "--backfill-leaderboard",
        action="store_true",
        help="backfill 2023 Q3 leaderboard data",
    )
    return parser


def shout_out_contribution(
    github_login: str,
    commit_sha: str,
    contribution_summary: str,
    check_url: str,
    is_backfill: bool,
) -> None:
    if not is_backfill and environ.get("GITHUB_EVENT_NAME", "") == "push":
        slack_handle = get_slack_handle_from_github_login(github_login)
        text = f"""\
🎉 Big Cheers for {slack_handle}! 🎉
A huge shout-out to {slack_handle} for the <https://github.com/Greenbax/evergreen/commit/{commit_sha}|{commit_sha[:7]}> change🌟 {contribution_summary}
Your contributions are now reflected on the Code Quality Award <{check_url}|Leaderboard>!👏
"""
        post_a_slack_message(TECH_DEBT_SHOUTOUT_CHANNEL, text)


async def analyze_contribution_and_create_a_check_run(  # noqa: PLR0917
    git_repo: Repo,
    github_repo: Repository,
    s3: S3Client,
    base_commit: Commit,
    head_commit: Commit,
    paths: list[str],
    analyzers: set[type[BaseCodeAnalyzer]],
    author: str,
    is_backfill: bool,
) -> None:
    (
        delta,
        path_results,
    ) = await PathResults.gen_from_incremental_analysis(
        s3,
        git_repo,
        base_commit,
        head_commit,
        paths,
        analyzers,
    )

    pr_comment_body = ""
    significant_contribution = ""
    pull = get_pull_request()
    try:
        try:
            leaderboard = await LeaderBoard.load_with_commit(s3, base_commit)
        except Exception:
            logger.warning("Failed to load leaderboard from s3.")
            leaderboard = LeaderBoard()
        try:
            history = await History.load_with_commit(s3, base_commit)
        except Exception:
            logger.warning("Failed to load history from s3.")
            history = History()
        summary = path_results.get_markdown_summary(
            delta, author, leaderboard.board.get(author, Counter[TechDebtCategory]())
        )
        leaderboard.add_delta_and_check_contribution(author, delta)
        leaderboard.compute_weighted_score()
        history.add_delta(
            author,
            head_commit.hexsha,
            delta,
            leaderboard.board[author][TechDebtCategory.WEIGHTED_SCORE],
        )
        try:
            await leaderboard.upload_to_s3(s3, head_commit)
        except Exception:
            logger.warning("Failed to upload leaderboard to s3.")
        try:
            await history.upload_to_s3(s3, head_commit)
        except Exception:
            logger.warning("Failed to upload history to s3.")
        summary += leaderboard.get_markdown_summary(author, 1000)
        summary += history.get_markdown_summary(author)
        if (
            leaderboard.tech_debt_regression
            and pull is not None
            and not has_previous_issue_comment(
                pull, RAISE_THE_BAR_LOGIN, NEW_TECH_DEBT_MESSAGE
            )
        ):
            pr_comment_body = f"Hey @{author}, {NEW_TECH_DEBT_MESSAGE}\n{leaderboard.tech_debt_regression}"
        if leaderboard.contribution_summary:
            significant_contribution = leaderboard.contribution_summary

    except ClientError:  # fail to load leaderboard
        summary = path_results.get_markdown_summary(delta, author)
        logger.warning("fail to load leaderboard or history")

    checks = create_check_run(
        repo=github_repo,
        name=CHECK_NAME,
        head_sha=head_commit.hexsha,
        conclusion=(
            "action_required"
            if any(
                val > 0
                for key, val in delta.items()
                if key
                not in ({
                    *REGRESSION_TECH_DEBT_CATEGORIES,
                    TechDebtCategory.WEIGHTED_SCORE,
                })
            )
            else "success"
        ),
        title="Python Tech Debt Report",
        summary=trim_to_max_bytes(summary, MAX_COMMENT_LENGTH),
        annotations=[],
        actions=[],
    )
    if pull and pr_comment_body:
        if checks:
            pr_comment_body += (
                f"\n\n[Check your Tech Debt Contribution]({checks[0].html_url})"
            )
        pull.create_issue_comment(pr_comment_body)
    if significant_contribution:
        shout_out_contribution(
            author,
            head_commit.hexsha,
            significant_contribution,
            checks[0].html_url,
            is_backfill,
        )


async def main() -> None:
    initialize_logging()
    args = get_parser().parse_args()
    analyzers = get_analyzers()
    logger.info(f"Analyzers: {analyzers}")
    session = Session()
    if args.backfill_leaderboard:
        git_repo = Repo()
        processed = 0
        github_repo = get_github_repo()
        # lint-fixme: NoS3ClientRule
        async with session.client(service_name="s3") as s3:  # pyright: ignore[reportUnknownMemberType]
            prev_commit = None
            for commit in list(git_repo.iter_commits())[::-1]:
                author = github_repo.get_commit(commit.hexsha).author.login
                logger.info(
                    f"{processed} {commit.hexsha[:7]} {commit.committed_datetime.isoformat()} [{author}]"
                )
                if prev_commit is None:
                    if commit.hexsha == args.start_hexsha:
                        prev_commit = commit
                    continue
                pr_url = f"https://github.com/Greenbax/evergreen/pull/{get_pr_num_from_commit_message(str(commit.summary))}"
                logger.info(pr_url)
                git_repo.git.checkout(commit)
                # lint-fixme: NoAwaitInLoopRule: this await cannot be gathered due to dependencies
                await analyze_contribution_and_create_a_check_run(
                    git_repo,
                    github_repo,
                    s3,
                    prev_commit,
                    commit,
                    args.paths,
                    analyzers,
                    author,
                    is_backfill=True,
                )
                processed += 1

                prev_commit = commit
    else:
        git_repo = get_git_repo()
        head_commit = git_repo.head.commit
        pull = get_pull_request()
        if pull:
            base_commit_sha = pull.get_commits()[0].parents[0].sha
        else:
            base_commit_sha = head_commit.parents[0].hexsha
        base_commit = git_repo.commit(base_commit_sha)
        # lint-fixme: NoS3ClientRule
        async with session.client("s3") as s3:  # pyright: ignore[reportUnknownMemberType]
            github_repo = get_github_repo()
            try:
                pull = get_pull_request()
                if pull is not None:
                    author = pull.user.login
                else:
                    author = github_repo.get_commit(head_commit.hexsha).author.login
            except AttributeError:
                author = "UnknownAuthor"
            await analyze_contribution_and_create_a_check_run(
                git_repo,
                github_repo,
                s3,
                base_commit,
                head_commit,
                args.paths,
                analyzers,
                author,
                is_backfill=False,
            )


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

from typing import TYPE_CHECKING

from fixit import InvalidTestCase as Invalid

from bar_raiser.tech_debt_framework.utils import BaseCodeAnalyzer, TechDebtCategory

if TYPE_CHECKING:
    from libcst import Comment


class FindPyrightIgnores(BaseCodeAnalyzer):
    INVALID = [  # noqa: RUF012
        Invalid("slack_text = subject  # pyright: ignore[reportUnboundVariable]")
    ]

    def visit_Comment(self, node: Comment) -> bool | None:
        if node.value.startswith("# pyright: ignore"):
            self.report(
                node,
                TechDebtCategory.PYRIGHT_IGNORE.value,
                "Pyright ignore comment found",
            )
        return super().visit_Comment(node)
